"""Tests for eigenvalue functionality."""

from matfree import decomp, eig
from matfree.backend import linalg, np, testing


@testing.parametrize("nrows", [7])
def test_equal_to_linalg_eig(nrows):
    """The output of full-depth decomposition should be equal (*) to linalg.svd().

    (*) Note: The singular values should be identical,
    and the orthogonal matrices should be orthogonal. They are not unique.
    """
    A = np.triu(np.arange(1.0, 1.0 + nrows**2).reshape((nrows, nrows)))
    v0 = np.ones((nrows,))
    num_matvecs = nrows

    hessenberg = decomp.hessenberg(num_matvecs, reortho="full")
    alg = eig.eig_partial(hessenberg)
    vals, vecs = alg(lambda v, p: p @ v, v0, A)

    S, U = linalg.eig(A)
    assert np.allclose(vecs.T @ vecs, U @ U.T, atol=1e-5, rtol=1e-5)
    assert np.allclose(vals, S)


@testing.parametrize("nrows", [8])
@testing.parametrize("num_matvecs", [8, 4, 0])
def test_shapes_as_expected_vector(nrows, num_matvecs):
    A = np.triu(np.arange(1.0, 1.0 + nrows**2).reshape((nrows, nrows)))

    v0 = np.ones((nrows,))
    v0 /= linalg.vector_norm(v0)

    hessenberg = decomp.hessenberg(num_matvecs, reortho="full")
    alg = eig.eig_partial(hessenberg)
    S, U = alg(lambda v, p: p @ v, v0, A)
    assert S.shape == (num_matvecs,)
    assert U.shape == (num_matvecs, nrows)


@testing.parametrize("nrows", [10])
@testing.parametrize("num_matvecs", [0, 2, 3])
def test_shapes_as_expected_lists_tuples(nrows, num_matvecs):
    K = np.arange(1.0, 10.0).reshape((3, 3))
    v0 = np.ones((nrows, nrows))  # tensor-valued input

    # Map Pytrees to Pytrees
    def Av(v: list, stencil) -> list:
        [(x,)] = v
        return [(np.convolve2d(x, stencil, mode="same"),)]

    hessenberg = decomp.hessenberg(num_matvecs, reortho="none")
    eig_partial = eig.eig_partial(hessenberg)
    vals, [(vecs,)] = eig_partial(Av, [(v0,)], K)

    # Ut inherits pytree-shape from the outputs (list),
    # and Vt inherits pytree-shape from the inputs (tuple)
    assert vecs.shape == (num_matvecs, *v0.shape)
    assert vals.shape == (num_matvecs,)



"""Tests for least-squares functionality."""

from matfree import lstsq, test_util
from matfree.backend import func, linalg, prng, testing
from matfree.backend.typing import Callable


@testing.case()
def case_lstsq_lsmr() -> Callable:
    return lstsq.lsmr(atol=1e-5, btol=1e-5, ctol=1e-5)


def case_A_shape_wide() -> tuple:
    return 3, 6


def case_A_shape_tall() -> tuple:
    return 6, 3


def case_A_shape_square() -> tuple:
    return 3, 3


@testing.parametrize_with_cases("lstsq_fun", cases=".", prefix="case_lstsq_")
@testing.parametrize_with_cases("A_shape", cases=".", prefix="case_A_shape_")
def test_value_and_grad_matches_numpy_lstsq(lstsq_fun: Callable, A_shape: tuple):
    key = prng.prng_key(1)

    key, subkey = prng.split(key, 2)
    matrix = prng.normal(subkey, shape=A_shape)
    key, subkey = prng.split(key, 2)
    rhs = prng.normal(subkey, shape=(A_shape[0],))
    key, subkey = prng.split(key, num=2)
    dsol = prng.normal(subkey, shape=(A_shape[1],))

    def lstsq_jnp(a, b):
        sol, *_ = linalg.lstsq(a, b)
        return sol

    expected, expected_vjp = func.vjp(lstsq_jnp, matrix, rhs)
    dmatrix1, drhs1 = expected_vjp(dsol)

    def vecmat(vector, p_as_list):
        [p] = p_as_list
        return p.T @ vector

    def lstsq_matfree(a, b):
        sol, _ = lstsq_fun(vecmat, a, b)
        return sol

    received, received_vjp = func.vjp(lstsq_matfree, rhs, [matrix])
    drhs2, [dmatrix2] = received_vjp(dsol)  # mind the order of rhs & matrix

    test_util.assert_allclose(received, expected)
    test_util.assert_allclose(dmatrix1, dmatrix2)
    test_util.assert_allclose(drhs1, drhs2)
